#!/usr/bin/perl -w

# $Id: rgraph.pl,v 1.4 2006/08/01 08:18:44 semik Exp $
#
# rgraph.pl is tool to graph SNMP Radiator stats
#
# Author: Jan Tomek <jan.tomasek@cesnet.cz>, <jan@tomasek.cz>
# Copyright (c) 2006, CESNET, z.s.p.o.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# See README and COPYING for more details.

use strict;
use RRDs;
use AppConfig qw(:expand :argcount);
use Net::SNMP;
use Data::Dumper;

# Global variables filed by info from config file
my $db = "";
my $servers;

# Global variables - sort of config
my $objects =
  {
   radiusAuthServTotalAccessAccepts           => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.8',
						  desc => "Access Accepts",
						  ord => "a02",
						  color => "#6a92b7",
						 },
   radiusAuthServTotalAccessRejects           => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.9',
						  desc => "Access Rejects",
						  ord => "a01",
						  color => "#ff6a60",
						 },
   radiusAuthServTotalAccessRequests          => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.5',
						  desc => "Access Requests",
						  ord => "a03",
						  color => "#c7ffae",
						 },
   radiusAuthServTotalAccessChallenges        => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.10',
						  desc => "Access Challenges",
						  ord => "a04",
						  color => "#80e454",
						 },
   radiusAuthServTotalUnknownTypes            => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.14',
						  desc => "Unknown Type Packets",
						  ord => "a05",
						  color => "#666666",
						 },
   radiusAuthServTotalInvalidRequests         => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.6',
						  desc => "Invalid Requests",
						  ord => "a06",
						  color => "#666666",
						 },
   radiusAuthServTotalMalformedAccessRequests => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.11',
						  desc => "Malformed Access Requests",
						  ord => "a07",
						  color => "#666666",
						 },
   radiusAuthServTotalDupAccessRequests       => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.7',
						  desc => "Duplicated Access Requests",
						  ord => "a08",
						  color => "#666666",
						 },
   radiusAuthServTotalPacketsDropped          => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.13',
						  desc => "Dropped Packets",
						  ord => "a09",
						  color => "#666666",
						 },
   radiusAuthServTotalBadAuthenticators       => {
						  mib => 'SNMPv2-SMI::mib-2.67.1.1.1.1.12',
						  desc => "Bad Authentiators",
						  ord => "a10",
						  color => "#666666",
						 },
  };

my $times = {
	         12  => {
			 desc => "12 hours",
			},
	       2*24  => {
			 desc => "2 days",
			},
	       7*24  => {
			 desc => "7 days",
			},
              31*24  => {
			 desc => "31 days",
			},
	     365*24  => {
			 desc => "1 year",
			},
	    };

# Global variables
my $now = time;

# I want to have date of graph creation in it's right corner, because
# I've no idea how to say rrdtool to put text into that place. I'm
# using spaces, that is not nice, but it works. You will maybe have to
# tune number of spaces to get it to right place depending on your
# font and date format.
#
# Situation get bit complicated because version above 1.2.11 (more
# likely, much more erlier) require escaping of : in comments. If you
# are getting error message:
#
# Error while creating graph: Garbage ':50:35 EDT 2006' after command:
#
# You need to tune version number below. Please let me known which
# version was causing troubles to you. This is mentioned in changelog
# from 2005-04-30 09:56. Version 1.2.0 was released at 2005-04-26 so I
# guess I'm right.

my $date = `date +"%d.%m.%Y %H:%M:%S"`; chomp($date);
my $signature = '<font size="-1"><p align="right">Created by <a href="http://wiki.eduroam.cz/rgraph">rgraph</a>.</p></font>'."\n";
my $edate = "                                                                                                      $date";
if ($RRDs::VERSION >= 1.2) {
  $edate = "                                                                                    $date";
  $edate =~ s/:/\\:/g;
};

# Global config.
my $config = AppConfig->new
  ({
    CASE => 1,
    GLOBAL=> {
	      EXPAND => EXPAND_ALL,
	      DEFAULT  => "<unset>",
	      ARGCOUNT => 1
	     },
   },
   CFGFilename         => {DEFAULT => './rgraph.cfg'},
   db                  => {DEFAULT => '/tmp'},
   createResizedGraphs => {DEFAULT => undef},
   server              => {ARGCOUNT => ARGCOUNT_LIST,
			   DEFAULT => undef},
  );

sub createDB {
  foreach my $server (keys %{$servers}) {
    mkdir "$db/$server" unless (-d "$db/$server");
    foreach my $id (keys %{$objects}) {
      next if (-f "$db/$server/$id.rrd");

      warn "Creating DB $server/$id\n";

      RRDs::create("$db/$server/$id.rrd",
		   "--step=1",
		   "--start=$now", # some time in past
		   "DS:value:ABSOLUTE:3600:0:U",
		   # 1600 values for 800px graph is enought
		   # 12 hours -> 12*60*60/1600 = 27
		   "RRA:MAX:0.5:27:1600",
		   "RRA:MIN:0.5:27:1600",
		   "RRA:AVERAGE:0.5:27:1600",
		   # 2 days -> 2*24*60*60/1600 = 108
		   "RRA:MAX:0.5:108:1600",
		   "RRA:MIN:0.5:108:1600",
		   "RRA:AVERAGE:0.5:108:1600",
		   # 8 dni -> 8*24*60*60/1600 = 432
		   "RRA:MAX:0.5:432:1600",
		   "RRA:MIN:0.5:432:1600",
		   "RRA:AVERAGE:0.5:432:1600",
		   # 70 dni -> 70*24*60*60/1600 = 3780
		   "RRA:MAX:0.5:3780:1600",
		   "RRA:MIN:0.5:3780:1600",
		   "RRA:AVERAGE:0.5:3780:1600",
		   # 370 dni -> 370*24*60*60/1600 = 19980
		   "RRA:MAX:0.5:19980:1600",
		   "RRA:MIN:0.5:19980:1600",
		   "RRA:AVERAGE:0.5:19980:1600",
		  );
      my $err = RRDs::error;
      die "Error while creating graph: $err\n" if $err;
    };
  };
};

sub updateDB {
  foreach my $server (keys %{$servers}) {
    my ($session, $error) =  Net::SNMP->session(
						-hostname   => $server,
						-version    => 1,
						-community  => $servers->{$server}->{comunity},
					       );
    unless ($session) {
      warn "Can't create SNMP session for $server: ".$error;
      next;
    };

    my $result = $session->get_request(-varbindlist => [
							map {$_ =~ s/SNMPv2-SMI::mib-2\./.1.3.6.1.2.1./; $_}
							map {$objects->{$_}->{mib}} keys %{$objects}
						       ]);
    unless ($result) {
      warn "Can't read SNMP data from $server: ".$session->error;
      next;
    };

    foreach my $id (keys %{$objects}) {
      my $oid = $objects->{$id}->{mib}; $oid =~ s/SNMPv2-SMI::mib-2\./.1.3.6.1.2.1./;
      #my $cmd = "snmpget -v 1 -c ".$servers->{$server}->{comunity}." $server ".$objects->{$id}->{mib};
      #my $no = `$cmd | sed "s/.*: //"`; chomp($no); my $value = $no;
      if (defined($result->{$oid})) {
	my $no = $result->{$oid}; my $value = $no;
	#my $last = `cat $db/$server/$id.last 2>/dev/null` || $no; chomp($last);
	my $last = $no;
	if (open(F, "<$db/$server/$id.last")) {
	  $last = <F> || $no; chomp($last);
	  close(F);
	};

	#`echo $no >$db/$server/$id.last`;
	open(F, ">$db/$server/$id.last") or die "Can't write to $db/$server/$id.last: $?";
	print F "$no\n";
	close(F);

	$value = $no - $last if ($no >= $last);
	
	if (open(F, ">>$db/$server/$id.txt")) {
	  print F "$now:$last:$no:$value\n";
	  close(F);
	};

	RRDs::update("$db/$server/$id.rrd",
		     "$now:$value");
	my $err = RRDs::error;
	warn "Error while updating $db/$server/$id.rrd: $err\n" if $err;
      };
    };
  };
};

my @DIMENSIONS=('--width=650',
		'--height=170');
my @colors = (
	      "--color=BACK#ffffff",
	      "--color=SHADEA#ffffff",
	      "--color=SHADEB#ffffff",
	      "--color=CANVAS#ffffff",
	      "--color=MGRID#777777",
	      "--color=FONT#074a87",
	      "--color=FRAME#074a87",
	      "--color=ARROW#074a87",
	     );

sub createGraph {
  foreach my $server (keys %{$servers}) {
    my @graph = map {"DEF:".$objects->{$_}->{ord}."=$db/$server/$_.rrd:value:AVERAGE"} (keys %{$objects});
    my @cdef;
    my @plot;
    my $no = 0;
    my @id;
    foreach my $id (sort {$objects->{$a}->{ord} cmp $objects->{$b}->{ord}} keys %{$objects}) {
      my $c = 0x111111*($no+2);

      if (@id) {
	push @cdef, "CDEF:b$no=".$objects->{$id}->{ord}.",".join(",+,",@id).',+';
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":MAX:max\\:%1.0lf %s [1/sec])\\n";
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":AVERAGE:avg\\:%1.2lf %s";
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":MIN:(min\\:%1.0lf %s";
	push @plot, "AREA:b$no".$objects->{$id}->{color}.":".$objects->{$id}->{desc}."";
      } else {
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":MAX:max\\:%1.0lf %s [1/sec])\\n";
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":AVERAGE:avg\\:%1.2lf %s";
	push @plot, "GPRINT:".$objects->{$id}->{ord}.":MIN:(min\\:%1.0lf %s";
	push @plot, "AREA:".$objects->{$id}->{ord}.$objects->{$id}->{color}.":".$objects->{$id}->{desc}."";
      };
      push @id, $objects->{$id}->{ord};
      $no++;
    };
    $no--;
    push @graph, @cdef;
    push @graph, reverse @plot;

    push @graph, "LINE1:b$no#074a87:Total packets";
    push @graph, "GPRINT:b$no:MIN:(min\\:%1.0lf %s";
    push @graph, "GPRINT:b$no:AVERAGE:avg\\:%1.2lf %s";
    push @graph, "GPRINT:b$no:MAX:max\\:%1.0lf %s [1/sec])\\n";

    my $htmlid = "";
    foreach my $time (sort {$a <=> $b} keys %{$times}) {
      my $start = `date +"%s" -d "$time hours ago"`; chomp($start);
      RRDs::graph("$db/$server/$time.gif",
		  "-z",
		  "-i",
		  "--start=$start",
		  "--end=$now",
		  '-l 0',
		  "-t Total Packets processed on $server in last ".$times->{$time}->{desc},
		  "-v Total Packets / sec",
		  @DIMENSIONS,
		  @colors,
		  @graph,
		  "COMMENT:$edate",
		 );
      my $err = RRDs::error;
      die "Error while creating graph: $err\n" if $err;

      if ($config->createResizedGraphs) {
	# TODO This should be done by calling perlmagick instead of
	# conver. It would be also nice if resize is called only if
	# gif file is never that png file.
	`convert $db/$server/$time.gif -resize x250 $db/$server/$time.png`;
      };

      $htmlid .= "<img src=\"$time.gif\"/><br/><br/>\n";
    };

    open(HTML, ">$db/$server/Total.html") or die "Can't write to \"$db/$server/Total.html\": $!";
    print HTML $htmlid;
    print HTML $signature;
    close(HTML);
  };
};

sub createGraphs {
  foreach my $server (keys %{$servers}) {
    my $html = "<a href=\"Total.html\"><img border=\"0\" src=\"48.gif\"/></a><br/><br/>\n";
    foreach my $id (keys %{$objects}) {
      my $htmlid = "";
      foreach my $time (sort {$a <=> $b} keys %{$times}) {
	my $start = `date +"%s" -d "$time hours ago"`; chomp($start);
	my $LEGEND = $objects->{$id}->{desc}."/sec";
	RRDs::graph("$db/$server/$id-$time.gif",
		    "-z",
		    "-i",
		    "--start=$start",
		    "--end=$now",
		    @DIMENSIONS,
		    @colors,
		    "-t ".$objects->{$id}->{desc}." on $server in last ".$times->{$time}->{desc},
		    "-v ".$objects->{$id}->{desc}." / sec",
		    "-l 0",
		    "DEF:avg=$db/$server/$id.rrd:value:AVERAGE",
		    "DEF:max=$db/$server/$id.rrd:value:MAX",
		    "DEF:min=$db/$server/$id.rrd:value:MIN",
		    "AREA:max#c1d1e1:min/max",
		    "LINE1:max#6a92b7:",
		    "AREA:min#ffffff:",
		    "LINE1:min#6a92b7:",
		    "GPRINT:min:MIN:(min\\:%1.0lf %s",
		    "GPRINT:max:MAX:max\\:%1.0lf %s $LEGEND)\\n",
		    "LINE2:avg#074a87:avg",
		    "GPRINT:avg:MIN:(min\\:%1.0lf %s",
		    "GPRINT:avg:AVERAGE:avg\\:%1.2lf %s",
		    "GPRINT:avg:MAX:max\\:%1.0lf %s $LEGEND)\\n",
		    "HRULE:0#000000",
		    "COMMENT:$edate",
		   );
	my $err = RRDs::error;
	die "Error while creating graph: $err\n" if $err;

	$htmlid .= "<img src=\"$id-$time.gif\"/><br/><br/>\n";
      };

      open(HTML, ">$db/$server/$id.html") or die "Can't write to \"$db/$server/$id.html\": $!";
      print HTML $htmlid;
      print HTML $signature;
      close(HTML);

      $html .= "<a href=\"$id.html\"><img border=\"0\" src=\"$id-48.gif\"/></a><br/><br/>\n";
    };

    open(HTML, ">$db/$server/index.html") or die "Can't write to \"$db/$server/index.html\": $!";
    print HTML $html;
    print HTML $signature;
    close(HTML);
  };
};

$config->args(\@ARGV) or
  die "Can't parse cmdline args";
$config->file($config->CFGFilename) or
  die "Can't open config file \"".$config->CFGFilename."\": $!";

$db = $config->db;
foreach my $srv (@{$config->server}) {
  if ($srv =~ /^(.*):(.*)$/) {
    $servers->{$1}->{comunity} = $2;
  } else {
    warn "Bad server definition expecting \"<servername>:<comunity>\". Got \"$srv\"";
  };
};

RRDs::graph

createDB();
updateDB();
createGraph();
createGraphs();
